/*******************************************************************
* SNAAB ILI9341 Arduino Library for SPI ILI9341 TFT Color Display
* version 1.0.0
* distributed under GPL v2.0
* Copyright (c) 2007-2021, SNAAB Systems 
* www.snaabsys.ru
*
* uses part of Stanley Huang (http://arduinodev.com) library code
*******************************************************************/

#include "Arduino.h"
#include <SPI.h>
#include "SNAAB_ILI9341.h"

/*************************************************************************
/* begin: Initializing the LCD display.
    CS    Pin of signal CS of Arduino
    DC    Pin of signal DC of Arduino
*/
/*************************************************************************/
void SNAAB_ILI9341::begin (uint8_t CS, uint8_t DC) {
	PIN_CS = CS;
	PIN_DC = DC;
    pinMode(PIN_CS, OUTPUT);
    pinMode(PIN_DC, OUTPUT);	
      SPI.begin();
      TFT_CS_HIGH;
    TFT_DC_HIGH;
    uint8_t i=0, TFTDriver=0;
    for(i=0;i<3;i++) {
      TFTDriver = readID();
    }
	sendCMD(0xCB);  
	  WRITE_DATA(0x39); 
	  WRITE_DATA(0x2C); 
	  WRITE_DATA(0x00); 
	  WRITE_DATA(0x34); 
	  WRITE_DATA(0x02); 
    sendCMD(0xCF);  
      WRITE_DATA(0x00); 
	  WRITE_DATA(0XC1); 
	  WRITE_DATA(0X30); 
	sendCMD(0xE8);  
	  WRITE_DATA(0x85); 
	  WRITE_DATA(0x00); 
	  WRITE_DATA(0x78); 
	sendCMD(0xEA);  
	  WRITE_DATA(0x00); 
	  WRITE_DATA(0x00); 
    sendCMD(0xED);  
	  WRITE_DATA(0x64); 
	  WRITE_DATA(0x03); 
	  WRITE_DATA(0X12); 
	  WRITE_DATA(0X81); 
	sendCMD(0xF7);  
	  WRITE_DATA(0x20); 
	sendCMD(0xC0);    	    // Power control 
	  WRITE_DATA(0x23);     // VRH[5:0] 
	sendCMD(0xC1);    	    // Power control 
	  WRITE_DATA(0x10);  	// SAP[2:0];BT[3:0] 
	sendCMD(0xC5);    	    // VCM control 
	  WRITE_DATA(0x3e);  	// Contrast
	  WRITE_DATA(0x28); 
	sendCMD(0xC7);    	    // VCM control2 
	  WRITE_DATA(0x86);     // --
	sendCMD(0x36);      	// Memory Access Control 
	  WRITE_DATA(0x48);  	// C8	   //48 68绔栧睆//28 E8 妯睆
	sendCMD(0x3A);    
	  WRITE_DATA(0x55); 
	sendCMD(0xB1);    
	  WRITE_DATA(0x00);  
	  WRITE_DATA(0x18); 
	sendCMD(0xB6);    	    // Display Function Control 
	  WRITE_DATA(0x08); 
	  WRITE_DATA(0x82);
	  WRITE_DATA(0x27);   
	sendCMD(0xF2);    	    // 3Gamma Function Disable 
	  WRITE_DATA(0x00); 
	sendCMD(0x26);    	    // Gamma curve selected 
	  WRITE_DATA(0x01); 
	sendCMD(0xE0);    	    // Set Gamma 
	  WRITE_DATA(0x0F); 
	  WRITE_DATA(0x31); 
	  WRITE_DATA(0x2B); 
	  WRITE_DATA(0x0C); 
	  WRITE_DATA(0x0E); 
	  WRITE_DATA(0x08); 
	  WRITE_DATA(0x4E); 
	  WRITE_DATA(0xF1); 
	  WRITE_DATA(0x37); 
	  WRITE_DATA(0x07); 
	  WRITE_DATA(0x10); 
	  WRITE_DATA(0x03); 
	  WRITE_DATA(0x0E); 
	  WRITE_DATA(0x09); 
	  WRITE_DATA(0x00); 
	sendCMD(0XE1);    	    // Set Gamma 
	  WRITE_DATA(0x00); 
	  WRITE_DATA(0x0E); 
	  WRITE_DATA(0x14); 
	  WRITE_DATA(0x03); 
	  WRITE_DATA(0x11); 
	  WRITE_DATA(0x07); 
	  WRITE_DATA(0x31); 
	  WRITE_DATA(0xC1); 
	  WRITE_DATA(0x48); 
	  WRITE_DATA(0x08); 
	  WRITE_DATA(0x0F); 
	  WRITE_DATA(0x0C); 
	  WRITE_DATA(0x31); 
	  WRITE_DATA(0x36); 
	  WRITE_DATA(0x0F); 
	sendCMD(0x11);    	    // Exit Sleep 
	  delay(120); 
	sendCMD(0x29);          // Display on 
	sendCMD(0x2c);   
      clear();
}

void SNAAB_ILI9341::clear(void) {
  setCol(0, MAX_X);
  setPage(0, MAX_Y);
  sendCMD(0x2c);                                                                                                                     
    TFT_DC_HIGH;
    TFT_CS_LOW;
  for(uint16_t i=0; i<38400; i++) {	    
    SPI.transfer(0);
    SPI.transfer(0);
    SPI.transfer(0);
    SPI.transfer(0);
  }
 TFT_CS_HIGH;
}

/*************************************************************************
/* setFreq: Sets the frequency of the SPI bus.
    devfreg    Divisor (2 - the maximum bus speed, 128 - minimum)
*/
/*************************************************************************/
void SNAAB_ILI9341::setFreq(int devfreq) {
  switch (devfreq){
    case 2:   SPI.setClockDivider(SPI_CLOCK_DIV2);   break;
    case 4:   SPI.setClockDivider(SPI_CLOCK_DIV4);   break; 
    case 8:   SPI.setClockDivider(SPI_CLOCK_DIV8);   break;
    case 16:  SPI.setClockDivider(SPI_CLOCK_DIV16);  break;
    case 32:  SPI.setClockDivider(SPI_CLOCK_DIV32);  break;
    case 64:  SPI.setClockDivider(SPI_CLOCK_DIV64);  break;
    case 128: SPI.setClockDivider(SPI_CLOCK_DIV128); break;	
  }  
}

/*************************************************************************
/* setOrientation: Sets the orientation display screen.
    val    If can take one of the folloving values:
           LANDSCAPE, PORTRAIT, LANDSCAPEROTATE, PORTRAITROTATE 
*/
/*************************************************************************/
void SNAAB_ILI9341::setOrientation(uint8_t val) {	
  Orientation = val;
}

/*************************************************************************
/* fillScreen: Fill a screen
    color Color in 16-bit format 
*/
/*************************************************************************/
void SNAAB_ILI9341::fillScreen(uint16_t color) {
  uint16_t x, y = 0;     
  switch (Orientation) {   
    case LANDSCAPE:       x = MAX_Y; y = MAX_X; break;
	case PORTRAIT:        x = MAX_X; y = MAX_Y; break;
	case LANDSCAPEROTATE: x = MAX_Y; y = MAX_X; break;
	case PORTRAITROTATE:  x = MAX_X; y = MAX_Y; break;
  } 
 doFillRect(0, x, 0, y, color);	  
}

/*************************************************************************
/* drawPixel: Draw a pixel
    X0    X coordinate
    Y0    Y coordinate
    color Color in 16-bit format 
*/
/*************************************************************************/
void SNAAB_ILI9341::drawPixel(uint16_t X0, uint16_t Y0, uint16_t color) {
  setXY(X0, X0, Y0, Y0);
  sendData(color);	
}

/*************************************************************************
/* drawHorLine: Draw a horizontal line
    XL      Leftmost X coordinate
    YL      Leftmost Y coordinate
    length  Line length
    color   Color in 16-bit format 
*/
/*************************************************************************/
void SNAAB_ILI9341::drawHorLine(uint16_t XL, uint16_t YL, uint16_t length, uint16_t color) { 
  if ((Orientation == LANDSCAPE) or (Orientation == PORTRAITROTATE)) {
    setXY(XL + length - 1, XL, YL, YL);	
  }
  if ((Orientation == PORTRAIT) or (Orientation ==  LANDSCAPEROTATE)) {
    setXY(XL, XL + length - 1, YL, YL);	
  }  
  for(int i=0; i<length; i++)
   sendData(color);
}

/*************************************************************************
/* drawVertLine: Draw a vertical line
    XL      Leftmost X coordinate
    YL      Leftmost Y coordinate
    length  Line length
    color   Color in 16-bit format 
*/
/*************************************************************************/
void SNAAB_ILI9341::drawVertLine(uint16_t XL, uint16_t YL, uint16_t length, uint16_t color) {
  if (Orientation == LANDSCAPEROTATE) {
	setXY(XL, XL, YL + length - 1, YL);  
  }
  else {
	setXY(XL, XL, YL, YL + length - 1);  
  }
  for(int i=0; i<length; i++)
   sendData(color);
}

/*************************************************************************
/* drawLine: Draw a line
    X0     The initial X coordinate
    Y0     The initial Y coordinate
    X1     The final X coordinate
    Y1     The final Y coordinate   
    color  Color in 16-bit format
*/
/*************************************************************************/
void SNAAB_ILI9341::drawLine(uint16_t X0, uint16_t Y0, uint16_t X1, uint16_t Y1, uint16_t color) {
  uint16_t  a, b, aY, x, y;
  int dx, dy;
  bool direct = false; 
  if (Y0 == Y1) {
    drawHorLine(X0, Y0, X1 - X0, color);
  }	  
  else if (X0 == X1) {
    drawVertLine(X0, Y0, Y1 - Y0, color);
  }	
  else {	
    direct = (((X0 < X1) && (Y0 > Y1)) || ((X0 > X1) && (Y0 < Y1)))?(true):(false);
	a = 0;
    aY = ((X0 > X1) && (Y0 > Y1))?(Y1):(Y0);
    if (X0 > X1) {
	 a = X0; 
	  X0 = X1;
	  X1 = a;			
	}
	if (Y0 > Y1) {  
	 a = Y0; 
	  Y0 = Y1;
	  Y1 = a;	  
	}		   
   dx = (X1 - X0);
   dy = (Y1 - Y0);	
    y = Y0;	
   b = (dy < dx)?(1):(0);
   for(x=X0; x<=X1; x++) {	 
     y = (direct == true)?(Y1 - ((dy * (x - X0 + b)) / dx)):(aY + ((dy * (x - X0 + b)) / dx)); 
	 drawPixel(x, y, color);  
   }
  } 	
}
	
/*************************************************************************
/* drawRect: Draws a rectangle with a thickness of one pixel
    XL     Leftmost X coordinate
    YL     Leftmost Y coordinate
    width  Rectangle width
    height Rectangle height  
    color  Color in 16-bit format
*/
/*************************************************************************/		
void SNAAB_ILI9341::drawRect(uint16_t XL, uint16_t YL, uint16_t width, uint16_t height, uint16_t color) {
  drawHorLine(XL, YL, width, color);
  drawHorLine(XL, YL + height, width +1, color);
  drawVertLine(XL, YL, height, color);
  drawVertLine(XL + width, YL, height, color);			
}

/*************************************************************************
/* drawThickRect: Draws a tick rectangle
    XL     Leftmost X coordinate
    YL     Leftmost Y coordinate
    width  Rectangle width
    height Rectangle height  
	tick   Rectangle tickness
    color  Color in 16-bit format
*/
/*************************************************************************/	

void SNAAB_ILI9341::drawThickRect(uint16_t XL, uint16_t YL, uint16_t width, uint16_t height, uint16_t thick, uint16_t color) {
  doFillRect(XL, XL + thick, YL, YL + height, color);	
  doFillRect(XL + width - thick, XL + width, YL, YL + height, color);	  
  doFillRect(XL + thick, XL + width - thick, YL, YL + thick, color);	
  doFillRect(XL + thick, XL + width - thick, YL + height - thick, YL + height, color);	
}

/*************************************************************************
/* drawRoundRect: Draws a rectangle with rounded corners one pixel thick
    XL     Leftmost X coordinate
    YL     Leftmost Y coordinate
    width  Rectangle width
    height Rectangle height  
    r      Radius of rounding corners
    color  Color in 16-bit format 
*/
/*************************************************************************/	
void SNAAB_ILI9341::drawRoundRect(uint16_t XL, uint16_t YL, uint16_t width, uint16_t height, uint16_t r,  uint16_t color) {
  uint16_t XC = 0;
  uint16_t YC = 0;
  drawHorLine(XL + r, YL, width - (2*r) +1, color);
  drawHorLine(XL + r, YL + height, width - (2*r) + 1, color);
  drawVertLine(XL, YL + r, height - (2*r) + 1, color);
  drawVertLine(XL + width, YL + r, height - (2*r) + 1, color);	        
  for(int i=0; i<4; i++){
    switch (i){
      case 0: XC = XL + r; YC = YL + r;                   break; 
      case 1: XC = XL + width -r; YC = YL + r;            break;  
      case 2: XC = XL + width - r; YC = YL + height - r;  break;  
      case 3: XC = XL + r; YC = YL + height - r;          break;      
    } 
   doDrawCorner(XC, YC, r, i, color);
  }  
}

/*************************************************************************
/* drawThickRoundRect: Draws a tick rectangle with rounded corners
    XL     Leftmost X coordinate
    YL     Leftmost Y coordinate
    width  Rectangle width
    height Rectangle height  
    r      Radius of rounding corners
	tick   Rectangle tickness
    color  Color in 16-bit format 
*/
/*************************************************************************/	

void SNAAB_ILI9341::drawThickRoundRect(uint16_t XL, uint16_t YL, uint16_t width, uint16_t height, uint16_t thick, uint16_t r,  uint16_t color) {
  uint16_t XC = 0;
  uint16_t YC = 0; 
  uint16_t d = r - thick; 
  doFillRect(XL, XL + thick, YL + r, YL + height - r, color);	
  doFillRect(XL + width - thick, XL + width, YL + r, YL + height - r, color);	  
  doFillRect(XL + r, XL + width - r, YL, YL + thick, color);	
  doFillRect(XL + r, XL + width - r, YL + height - thick, YL + height, color);		

  for(int i=0; i<4; i++){
    switch (i){
      case 0: XC = XL + r; YC = YL + r;                   break; 
      case 1: XC = XL + width -r; YC = YL + r;            break;  
      case 2: XC = XL + width - r; YC = YL + height - r;  break;  
      case 3: XC = XL + r; YC = YL + height - r;          break;      
    } 
   doDrawThickCorner(XC, YC, r, d, i, color);
  }
}

/*************************************************************************
/* fillRect: Draws a filled rectangle
    XL     Leftmost X coordinate
    YL     Leftmost Y coordinate
    width  Rectangle width
    height Rectangle height  
    color  Color in 16-bit format
*/
/*************************************************************************/	
void SNAAB_ILI9341::fillRect(uint16_t XL, uint16_t YL, uint16_t width, uint16_t height, uint16_t color) {
  doFillRect(XL, XL + width, YL, YL + height, color);   
}

/*************************************************************************
/* fillRoundRect: Draws a filled rectangle with rounded corners
    XL     Leftmost X coordinate
    YL     Leftmost Y coordinate
    width  Rectangle width
    height Rectangle height  
    r      Radius of rounding corners
    color  Color in 16-bit format 
*/
/*************************************************************************/	
void SNAAB_ILI9341::fillRoundRect(uint16_t XL, uint16_t YL, uint16_t width, uint16_t height, uint16_t r, uint16_t color) {
  uint16_t XC = 0;
  uint16_t YC = 0;
  doFillRect(XL, XL + width, YL + r, YL + height - r, color);                // central unit
  doFillRect(XL + r, XL + width - r, YL, YL + r, color);                     // upper block 
  doFillRect(XL + r, XL + width - r, YL + height - r, YL + height, color);   // lower block
  for(int i=0; i<4; i++){
    switch (i){
      case 0: XC = XL + r; YC = YL + r;                   break; 
      case 1: XC = XL + width -r; YC = YL + r;            break;  
      case 2: XC = XL + width - r; YC = YL + height - r;  break;  
      case 3: XC = XL + r; YC = YL + height - r;          break;      
    } 
   doFillCorner(XC, YC, r, i, color);
  }      
}

/*************************************************************************
/* drawCircle: Draws a circle with a thickness of one pixel
    X0     The X coordinate of the center of the circle
    Y0     The Y coordinate of the center of the circle
    r      Circle radius
    color  Color in 16-bit format 
*/
/*************************************************************************/	
void SNAAB_ILI9341::drawCircle(uint16_t X0, uint16_t Y0, uint16_t r, uint16_t color) {
   uint16_t XC = 0;
   uint16_t YC = 0;   
   for(int i=0; i<4; i++){	
    switch (i){
      case 0: XC = X0 + 1; YC = Y0 + 1;  break; 
      case 1: XC = X0 - 1; YC = Y0 + 1;  break;  
      case 2: XC = X0 - 1; YC = Y0 - 1;  break;  
      case 3: XC = X0 + 1; YC = Y0 - 1;  break;      
    }   
    doDrawCorner(XC, YC, r + 1, i, color);
   }	
}

/*************************************************************************
/* fillCircle: Draws a filled circle
    X0     The X coordinate of the center of the circle
    Y0     The Y coordinate of the center of the circle
    r      Circle radius
    color  Color in 16-bit format
*/
/*************************************************************************/	
void SNAAB_ILI9341::fillCircle(uint16_t X0, uint16_t Y0, uint16_t r, uint16_t color) {
  uint16_t xx; 
  uint16_t yy; 
    xx = sqrt((r * r)/2);
    yy = xx;  
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r;
  while (x < y) {	  
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;	
    doFillRect(X0 - y -1, X0 - y, Y0 - x, Y0 - x + 2 * x - 1, color);	// left segment
	doFillRect(X0 - x, X0 - x + 2 * x +1, Y0 - y - 1, Y0 - y, color);	// upper segment 
	doFillRect(X0 + y, X0 + y +1, Y0 - x -1, Y0 - x + 2 * x, color);    // right segment 
	doFillRect(X0 + x - 2 * x + 1, X0 + x, Y0 + y, Y0 + y +1, color);	// lower segment
  }	
 fillRect(X0 - xx, Y0 - yy, 2 * xx, 2 * yy, color);                    // the central square
}           

/*************************************************************************
/* drawRing: Draws a tick ring
    X0     The X coordinate of the center of the ring
    Y0     The Y coordinate of the center of the ring
    r      Reng radius
	thick  Ring tickness
    color  Color in 16-bit format
*/
/*************************************************************************/	

void SNAAB_ILI9341::drawRing(uint16_t X0, uint16_t Y0, uint16_t r, uint16_t thick, uint16_t color) {
  uint16_t XC = 0;
  uint16_t YC = 0; 
  uint16_t d = r - thick; 	  
  for(int i=0; i<4; i++){	
   switch (i){
     case 0: XC = X0 + 1; YC = Y0 + 1;  break; 
     case 1: XC = X0 - 1; YC = Y0 + 1;  break;  
     case 2: XC = X0 - 1; YC = Y0 - 1;  break;  
     case 3: XC = X0 + 1; YC = Y0 - 1;  break;      
   }   
   doDrawThickCorner(XC, YC, r, d, i, color);
  }  	
}	

/*************************************************************************
/* setFont: Sets the current font
    font   The CurrentFont object
*/
/*************************************************************************/	
void SNAAB_ILI9341::setFont(const CurrentFont *font) {
  if (cFont != font) {	
    cFont = (CurrentFont *)font; 
	  fcharcount = pgm_read_byte(&cFont->charcount); 
	  fwidth = pgm_read_word(&cFont->width);
	  fheight = pgm_read_word(&cFont->height); 
      fascent = pgm_read_word(&cFont->ascent);
    lChar = pgm_read_word((ListChar *)&cFont->avchar); 
    cInfo = pgm_read_word((CharInfo *)&cFont->info);	
    fData = pgm_read_word((uint8_t *)&cFont->data);	
  }
}

/*************************************************************************
/* drawChar: Draws a single character
    ct         The 16-bit character code
    X0         The initial X coordinate
    Y0         The initial Y coordinate
    baseline   Binding to the font baseline. If the baseline = true, the 			   
			   X0 - is the leftmost X coordinate, Y0 is the Y coordinate 
			   the font baseline, otherwise X0 and Y0 coordinates define 
			   the upper-left corner of the character
    color      Color in 16-bit format 
	return     If the character is missing, returns = the font width, 
	           otherwise - the sum of the actual width and the right 
			   indentation of the character in pixels 
	Attention! Before using it, be sure to set the font with the setFont	
*/
/*************************************************************************/	
uint16_t SNAAB_ILI9341::drawChar(uint16_t ct, int16_t X0, int16_t Y0, bool baseline, uint16_t color) {
  uint8_t   a, b, c, d = 0;	  
  uint8_t   i, x, y = 0;
  unsigned char   cct;
  uint8_t   topoffset = 0;
  uint8_t   rwidth = 0;
  uint8_t   space = 0;  
  uint8_t   bsize = 0;
  uint16_t  bstart = 0; 
// UTF-8 to Win1251 
  cct = UTF8to1251(ct);  
// determint the presence of the symbol
  a = cct / lcStep;
  b = lcStep - (cct - (a * lcStep)) -1;
  c = pgm_read_byte(&lChar->listofchar[a]);
  d = bitRead(c, b);
  if (baseline == true) {
    Y0 = Y0 - fascent;	 
  } 
// if there is no symbol, draw a contour
  if (d == 0) {
 	drawRect(X0, Y0 +1, fwidth, fheight, color); 
  }	 
  else {
// if there is a symbol, draw
	for(uint8_t i=0; i<fcharcount; i++){	
	 a = pgm_read_byte(&cInfo[i].code);	
	 if (a == cct) {	   
	  topoffset = pgm_read_byte(&cInfo[i].topoffset);
	  rwidth = pgm_read_byte(&cInfo[i].realwidth);
	  space = pgm_read_byte(&cInfo[i].space);	  
	  bsize = pgm_read_word(&cInfo[i].blocksize);
	  bstart = pgm_read_word(&cInfo[i].blockstart);	  
	  break;  
	 }	
	} 
    a = b = c = x = y = 0;
    for (i = 0; i < bsize; i++) {
      a = pgm_read_byte(&fData[bstart + i]);  
	  for (b = 0; b < lcStep; b++) {			 
		c = bitRead(a, (lcStep - b -1)); 	
		x++;
        if (c == 1) {
	     drawPixel(X0 + x, Y0 + y + topoffset +1, color); 		  
		}  	   
		if (x >= rwidth) {
	     x = 0;		
         y++;		 
		}	   
	  }  
	}  
  } 
 if (d == 0) {
  return fwidth;	 
 }
 else {
  return rwidth + space;	 
 }
}

/*************************************************************************
/* drawChar: Draws a single character
    font       The CurrentFont object
    ct         The 16-bit character code
    X0         The initial X coordinate
    Y0         The initial Y coordinate
    baseline   Binding to the font baseline. If the baseline = true, the 			   
			   X0 - is the leftmost X coordinate, Y0 is the Y coordinate 
			   the font baseline, otherwise X0 and Y0 coordinates define 
			   the upper-left corner of the character
    color      Color in 16-bit format
	return     If the character is missing, returns = the font width, 
	           otherwise - the sum of the actual width and the right 
			   indentation of the character in pixels 
*/
/*************************************************************************/	
uint16_t SNAAB_ILI9341::drawChar(const CurrentFont *font, uint16_t ct, int16_t X0, int16_t Y0, bool baseline, uint16_t color) {
  setFont(font);
  return drawChar(ct, X0, Y0, baseline, color);
}

/*************************************************************************
/* drawChar: Draws a string of characters
    str        Character string
    X0         The initial X coordinate
    Y0         The initial Y coordinate
    baseline   Binding to the font baseline. If the baseline = true, the 			   
			   X0 - is the leftmost X coordinate, Y0 is the Y coordinate 
			   the font baseline, otherwise X0 and Y0 coordinates define 
			   the upper-left corner of the character
	mode	   Return value modifier. If mode = true, then return = count 
	           characters of string, otherwise return = the length of the 
			   string in pixels
    color      Color in 16-bit format
	return     See mode
	Attention! Before using it, be sure to set the font with the setFont
*/
/*************************************************************************/	
uint16_t SNAAB_ILI9341::drawText(char *str, int16_t X0, int16_t Y0, bool baseline, bool mode, uint16_t color) {
  uint16_t   i, k, n = 0; 
  unsigned char   cct;
  String    Str1251;
// UTF-8 to Win1251
  Str1251 = UTF8to1251(str); 
// output the characters
  n = 0;
  for (i = 0; i < (Str1251.length()); i++) {
   cct = Str1251[i];	
    k = drawChar(cct, X0 + n, Y0, baseline, color);
   n += k;
  }
  if (mode == true) {
   return i;
  }
  else {
   return n;
  }  
}	

/*************************************************************************
/* drawText: Draws a string of characters
    font       The CurrentFont object
    str        Character string
    X0         The initial X coordinate
    Y0         The initial Y coordinate
    baseline   Binding to the font baseline. If the baseline = true, the 			   
			   X0 - is the leftmost X coordinate, Y0 is the Y coordinate 
			   the font baseline, otherwise X0 and Y0 coordinates define 
			   the upper-left corner of the character
	mode	   Return value modifier. If mode = true, then return = count 
	           characters of string, otherwise return = the length of the 
			   string in pixels				
    color      Color in 16-bit format
	return     See mode	
*/
/*************************************************************************/	
uint16_t SNAAB_ILI9341::drawText(const CurrentFont *font, char *str, int16_t X0, int16_t Y0, bool baseline, bool mode, uint16_t color) {
  setFont(font);	
  return drawText(str, X0, Y0, baseline, mode, color);	  
}

/*************************************************************************
/* fillChar: Fills in the familiar character
    str        Character string
    X0         The initial X coordinate
    Y0         The initial Y coordinate
    baseline   Binding to the font baseline. If the baseline = true, the 			   
			   X0 - is the leftmost X coordinate, Y0 is the Y coordinate 
			   the font baseline, otherwise X0 and Y0 coordinates define 
			   the upper-left corner of the character
	index	   The number of the character in the string 	   
    color      Color in 16-bit format
	Attention! Before using it, be sure to set the font with the setFont
*/
/*************************************************************************/	
void SNAAB_ILI9341::fillChar(char *str, int16_t X0, int16_t Y0, bool baseline, uint16_t index, uint16_t color) {
  uint8_t   a, b, c, d = 0;	  
  uint8_t   i, k, x, y = 0;
  unsigned char cct;
  String    Str1251;
  uint8_t   topoffset;
  uint8_t   rwidth;
  uint8_t   rheight;
  uint8_t   space; 
  uint8_t   xwidth = 0;
// UTF-8 to Win1251
  Str1251 = UTF8to1251(str); 
  if (baseline == true) {
    Y0 = Y0 - fascent;	 
  } 
// output the characters
  xwidth = 0;
  for (k = 0; k < (Str1251.length()); k++) {
   cct = Str1251[k];	
 // determint the presence of the symbol 
  a = cct / lcStep;
  b = lcStep - (cct - (a * lcStep)) -1;
  c = pgm_read_byte(&lChar->listofchar[a]);
  d = bitRead(c, b);    
// if there is no symbol, skip it
  if (d == 0) {
	xwidth = xwidth + fwidth + (fwidth/7);
  }	 
  else {
// it there is a symbol, fill in the rectangle
	for(uint8_t i=0; i<fcharcount; i++){	
	 a = pgm_read_byte(&cInfo[i].code);	
	 if (a == cct) {	   
	  topoffset = pgm_read_byte(&cInfo[i].topoffset);
	  rwidth = pgm_read_byte(&cInfo[i].realwidth);
	  rheight = pgm_read_byte(&cInfo[i].realheight);
	  space = pgm_read_byte(&cInfo[i].space);
	  break;  
	 }	
	} 	
	if (index == k) {	
      a = X0 + xwidth;
	  b = Y0 + (topoffset/2) +1;
	  c = rwidth +(space/2);
	  d = rheight +(topoffset/2) +1;	  
      fillRect(a, b, c, d, color);	
	}
   xwidth = xwidth + rwidth + space;
  }
 }	
}

/*************************************************************************
/* fillChar: Fills in the familiar character
    str        Character string
    X0         The initial X coordinate
    Y0         The initial Y coordinate			   
    baseline   Binding to the font baseline. If the baseline = true, the 			   
			   X0 - is the leftmost X coordinate, Y0 is the Y coordinate 
			   the font baseline, otherwise X0 and Y0 coordinates define 
			   the upper-left corner of the character			   
	index	   The number of the character in the string 	   
    color      Color in 16-bit format
*/
/*************************************************************************/	
void SNAAB_ILI9341::fillChar(const CurrentFont *font, char *str, int16_t X0, int16_t Y0, bool baseline, uint16_t index, uint16_t color) {
  setFont(font);	
  fillChar(str, X0, Y0, baseline, index, color);	
}

/*************************************************************************
/* drawBitmap: Draw PROGMEM-resident bitmap files (*.h), created by 
               SNAAB FontEditor. For more details, see 
               SNAAB FontEditor (www.snaabsys.ru).
    XL     The left X coordinate
    YL     The top Y coordinate
    w      Width of bitmap in pixels
    h      Height of bitmap in pixels
    color  Color in 16-bit format
*/
/*************************************************************************/
void SNAAB_ILI9341::drawBitmap(const uint8_t bitmap[], int16_t XL, int16_t YL, int16_t w, int16_t h, uint16_t color) {
  uint8_t byte, n, m = 0;
  int16_t lsize, a, k, t = 0;
  int16_t x, y;
  x = 0;
  y = 0;
  t = w * h;
  lsize = t / lcStep;	
    k = (t > (lsize * lcStep))?(1):(0);
  lsize += k;  
   for (a = 0; a < lsize; a++) {	   
// we read one byte from memory
     byte = pgm_read_byte(&bitmap[a]);	   
// we read every bit in a byte   
     for (m = 0; m < lcStep; m++) {
       n = bitRead(byte, lcStep - m -1);
       if (n > 0) drawPixel(XL + x, YL + y, color); 	
		++x;	
       if (x > (w - 1)) {
        x = 0; 
        ++y;		
	   }  
     }    	   
   } 
}	
		
void SNAAB_ILI9341::sendCMD(uint8_t index) {
  TFT_DC_LOW;
  TFT_CS_LOW;
  SPI.transfer(index);
  TFT_CS_HIGH;
}

void SNAAB_ILI9341::setCol(uint16_t StartCol,uint16_t EndCol) {
  sendCMD(0x2A);                            
  sendData(StartCol);
  sendData(EndCol);
}

void SNAAB_ILI9341::setPage(uint16_t StartPage,uint16_t EndPage) {
  sendCMD(0x2B);                           
  sendData(StartPage);
  sendData(EndPage);
}

void SNAAB_ILI9341::WRITE_DATA(uint8_t data) {
  TFT_DC_HIGH;
  TFT_CS_LOW;
  SPI.transfer(data);
  TFT_CS_HIGH;
}

void SNAAB_ILI9341::sendData(uint16_t data) {
  uint8_t data1 = data>>8;
  uint8_t data2 = data&0xff;
  TFT_DC_HIGH;
  TFT_CS_LOW;
  SPI.transfer(data1);
  SPI.transfer(data2);
  TFT_CS_HIGH;
}

uint8_t SNAAB_ILI9341::Read_Register(uint8_t Addr, uint8_t xParameter) {
  uint8_t data=0;
  sendCMD(0xd9);                                                     
  WRITE_DATA(0x10+xParameter);                                       
  TFT_DC_LOW;
  TFT_CS_LOW;
  SPI.transfer(Addr);
  TFT_DC_HIGH;
  data = SPI.transfer(0);
  TFT_CS_HIGH;
  return data;
}

uint8_t SNAAB_ILI9341::readID(void) {
  uint8_t i=0;
  uint8_t data[3] ;
  uint8_t ID[3] = {0x00, 0x93, 0x41};
  uint8_t ToF=1;	
  for(i=0;i<3;i++) {
    data[i]=Read_Register(0xd3,i+1);
     if(data[i] != ID[i]) {
       ToF=0;
     }
  }
 return ToF;
}

void SNAAB_ILI9341::setXY(uint16_t x0, uint16_t x1, uint16_t y0, uint16_t y1) {   
  switch (Orientation) {   		
	case LANDSCAPE:       setCol(y0, y1); setPage(MAX_Y - x0, MAX_Y - x1);                  break;		
	case PORTRAIT:        setCol(x0, x1); setPage(y0, y1);                                  break; 
	case LANDSCAPEROTATE: setCol(MAX_X - y0, MAX_X - y1); setPage(x0, x1);                  break; 
	case PORTRAITROTATE:  setCol(MAX_X - x0, MAX_X - x1); setPage(MAX_Y - y1, MAX_Y - y0);  break; 
  } 
 sendCMD(0x2c);
}

void SNAAB_ILI9341::doFillRect(uint16_t XL, uint16_t XR, uint16_t YT, uint16_t YB, uint16_t color) {
  uint8_t Hcolor = color>>8;
  uint8_t Lcolor = color&0xff;
  unsigned long  n = 0;
  unsigned long i = 0;  
  switch (Orientation) {   
	case LANDSCAPE:       XR = min(XR, MAX_Y); YB = min(YB, MAX_X); setXY(XR, XL, YT, YB);  break;	
	case PORTRAIT:        XR = min(XR, MAX_X); YB = min(YB, MAX_Y); setXY(XL, XR, YT, YB);  break;
	case LANDSCAPEROTATE: XR = min(XR, MAX_Y); YB = min(YB, MAX_X); setXY(XL, XR, YB, YT);  break; 
	case PORTRAITROTATE:  XR = min(XR, MAX_X); YB = min(YB, MAX_Y); setXY(XR, XL, YT, YB);  break; 		
  }   
  TFT_DC_HIGH;
  TFT_CS_LOW;  
  n = XR - XL + 1;
  n = n * (YB - YT + 1);  
  for(i = n; i > 0; i--) {	  	  
    SPI.transfer(Hcolor);
    SPI.transfer(Lcolor);
  }
 TFT_CS_HIGH;
}

void SNAAB_ILI9341::doDrawCorner(uint16_t X0, uint16_t Y0, uint16_t r, uint8_t a, uint16_t color) {  
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r;
  while (x < y) {	  
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;	
   switch (a) {
     case 0: drawPixel(X0 - x, Y0 - y, color); drawPixel(X0 - y, Y0 - x, color);  break;
     case 1: drawPixel(X0 + x, Y0 - y, color); drawPixel(X0 + y, Y0 - x, color);  break;  
     case 2: drawPixel(X0 + x, Y0 + y, color); drawPixel(X0 + y, Y0 + x, color);  break;  
     case 3: drawPixel(X0 - y, Y0 + x, color); drawPixel(X0 - x, Y0 + y, color);  break; 
   }		
  }	 
}

void SNAAB_ILI9341::doDrawThickCorner(uint16_t X0, uint16_t Y0, uint16_t r, uint16_t delta, uint8_t a, uint16_t color) {
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r; 
  int16_t ft = 1 - delta;
  int16_t ddF_xt = 1;
  int16_t ddF_yt = -2 * delta;
  int16_t xt = 1;
  int16_t yt = delta;  
  if ((a ==0) or (a ==1)) {
   yt -= 1;
  }	 
  while (x < y) {	  
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;		
	if (xt < yt) {		
      if (ft >= 0) {
        yt--;
        ddF_yt += 2;
        ft += ddF_yt;
      }
      xt++;
      ddF_xt += 2;
      ft += ddF_xt;			
	}
   switch (a) {
     case 0: drawVertLine(X0 - x, Y0 - y, y - yt, color); drawHorLine(X0 - y, Y0 - x, y - yt, color); break;		 			 
     case 1: drawVertLine(X0 + x, Y0 - y, y - yt, color); drawHorLine(X0 + yt +1, Y0 - x, y - yt, color); break;			 
     case 2: drawVertLine(X0 + x, Y0 + yt, y - yt +1, color); drawHorLine(X0 + yt, Y0 + x, y - yt +1, color); break;			 			 
     case 3: drawVertLine(X0 - x, Y0 + yt, y - yt +1, color); drawHorLine(X0 - y, Y0 + x, y - yt +1, color); break;				 
   }		
  } 
}

void SNAAB_ILI9341::doFillCorner(uint16_t X0, uint16_t Y0, uint16_t r, uint8_t a, uint16_t color) {
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r;
  while (x < y) {	  
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;	
   switch (a){
	 case 0: drawVertLine(X0 - x, Y0 - y, y, color); drawVertLine(X0 - y, Y0 - x, x, color);  break;
     case 1: drawVertLine(X0 + x, Y0 - y, y, color); drawVertLine(X0 + y, Y0 - x, x, color);  break;	 	 
	 case 2: drawVertLine(X0 + x, Y0, y, color); drawVertLine(X0 + y, Y0, x, color);          break;
	 case 3: drawVertLine(X0 - y, Y0, x, color); drawVertLine(X0 - x, Y0, y, color);          break; 
   }		
  }	   
}

String SNAAB_ILI9341::UTF8to1251(char *str) {
  uint8_t  i = 0;	
  int  n = 0; 	
  unsigned char  cct = 0;
  char  m[2] = {'0', '\0'};
  String  Str1251;  
// UTF8 to Win1251
  n = strlen(str); 
  i = 0;   
  while (i < n) {	 
   cct = str[i]; 
   i++;  
   if (cct > 0x7F) {
     switch (cct) {	 
	   case 0xE2: { 
	    i++;
	   }	 
       case 0x84: { 
        cct = str[i]; 
        i++;		 
// №		
        if (cct == 0x96) { 	   
          cct = 0xB9;  break;         
        }
       }			   
       case 0xD0: {            			  
        cct = str[i]; 
        i++;		  
// Ё			  
       if (cct == 0x81) { 	   
         cct = 0xA8;  break;         
       }	 
// the resr of the Cyrillic alphabet		  
       if (cct >= 0x90 && cct <= 0xBF) {
         cct = cct + 0x30; break;          
       }                       
      }			
      case 0xD1: {
       cct = str[i]; 
	   i++;				 
// ё	   
       if (cct == 0x91) { 	   
         cct = 0xB8; break;          
       }		  
// the rest of the Cyrillic alphabet	  
       if (cct >= 0x80 && cct <= 0x8F) {
         cct = cct + 0x70; break;         
       }          
      }  
    }	
   }	 
   m[0] = cct; 
   Str1251 = Str1251 + String(m);
  } 
 return Str1251;
}

unsigned char SNAAB_ILI9341::UTF8to1251(uint16_t ct) {
  uint8_t  a = 0;	
  unsigned char  cct = 0;  
// UTF8 to Win1251
  cct = ct;
  if (ct > 0x7F) {
   a = highByte(ct);
    switch (a) {
      case 0x84: { 	  
// №		
       if (cct == 0x96) { 	   
         cct = 0xB9;  break;         
       }
	  }	
      case 0xD0: {         
       cct = lowByte(ct);
// Ё	   
       if (cct == 0x81) { 	   
         cct = 0xA8;  break;         
       }
        if (cct >= 0x90 && cct <= 0xBF) {
          cct = cct + 0x30;  break;            
        }                       
      }
      case 0xD1: {
       cct = lowByte(ct);
// ё	   
       if (cct == 0x91) { 	   
         cct = 0xB8;  break;          
       }
// the rest of the Cyrillic alphabet		   
        if (cct >= 0x80 && cct <= 0x8F) {
          cct = cct + 0x70;  break;         
        }          
      }        
    }	
  }	 
 return cct;
}
